'''
UHPPOTE request packet encoder.

Encodes a UHPPOTE access controller request as a 64 byte UDP packet:

- uint8, uint16, uint24 and uint32 values are encoded as little endian unsigned integers
- datetime, date and time values are encoded as BCD
- boolean values are encoded as 0 (False) or 1 (True)
'''

import struct


def get_controller_request(controller):
    '''
    Encodes a get-controller request.

        Parameters:
            controller (uint32)  Controller serial number.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x94

    pack_uint32(controller, packet, 4)

    return packet


def set_ip_request(controller, address, netmask, gateway):
    '''
    Encodes a set-ip request.

        Parameters:
            controller (uint32)       Controller serial number.
            address    (IPv4Address)  Controller IP address.
            netmask    (IPv4Address)  Controller subnet mask.
            gateway    (IPv4Address)  Controller gateway address.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x96

    pack_uint32(controller, packet, 4)
    pack_IPv4(address, packet, 8)
    pack_IPv4(netmask, packet, 12)
    pack_IPv4(gateway, packet, 16)
    pack_uint32(0x55aaaa55, packet, 20)

    return packet


def get_time_request(controller):
    '''
    Encodes a get-time request.

        Parameters:
            controller (uint32)  Controller serial number.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x32

    pack_uint32(controller, packet, 4)

    return packet


def set_time_request(controller, datetime):
    '''
    Encodes a set-time request.

        Parameters:
            controller (uint32)    Controller serial number.
            datetime   (datetime)  Date and time.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x30

    pack_uint32(controller, packet, 4)
    pack_datetime(datetime, packet, 8)

    return packet


def get_status_request(controller):
    '''
    Encodes a get-status request.

        Parameters:
            controller (uint32) Controller serial number.
 
        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x20

    pack_uint32(controller, packet, 4)

    return packet


def get_listener_request(controller):
    '''
    Encodes a get-listener request.

        Parameters:
            controller (uint32) Controller serial number.
 
        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x92

    pack_uint32(controller, packet, 4)

    return packet


def set_listener_request(controller, address, port):
    '''
    Encodes a set-listener request.

        Parameters:
            controller (uint32)                 Controller serial number.
            address    (ipaddress.IPv4Address)  IP address of event listener.
            port       (uint16)                 UDP port of event listener.
 
        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x90

    pack_uint32(controller, packet, 4)
    pack_IPv4(address, packet, 8)
    pack_uint16(port, packet, 12)

    return packet


def get_door_control_request(controller, door):
    '''
    Encodes a get-door-control request.

        Parameters:
            controller (uint32) Controller serial number.
            door       (uint8)  Door ID [1..4]
 
        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x82

    pack_uint32(controller, packet, 4)
    pack_uint8(door, packet, 8)

    return packet


def set_door_control_request(controller, door, mode, delay):
    '''
    Encodes a set-door-control request.

        Parameters:
            controller (uint32) Controller serial number.
            door       (uint8)  Door ID [1..4]
            mode       (uint8)  Control mode (1: normally open, 2: normally closed, 3: controlled)
            delay      (uint8)   Door unlock duration (seconds)

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x80

    pack_uint32(controller, packet, 4)
    pack_uint8(door, packet, 8)
    pack_uint8(mode, packet, 9)
    pack_uint8(delay, packet, 10)

    return packet


def open_door_request(controller, door):
    '''
    Encodes an open-door request.

        Parameters:
            controller (uint32) Controller serial number.
            door       (uint8)  Door ID [1..4]

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x40

    pack_uint32(controller, packet, 4)
    pack_uint8(door, packet, 8)

    return packet


def get_cards_request(controller):
    '''
    Encodes a get-cards request.

        Parameters:
            controller (uint32) Controller serial number.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x58

    pack_uint32(controller, packet, 4)

    return packet


def get_card_request(controller, card_number):
    '''
    Encodes a get-card request.

        Parameters:
            controller  (uint32) Controller serial number.
            card_number (uint32) Card number.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x5a

    pack_uint32(controller, packet, 4)
    pack_uint32(card_number, packet, 8)

    return packet


def get_card_by_index_request(controller, card_index):
    '''
    Encodes a get-card-by-index request.

        Parameters:
            controller  (uint32) Controller serial number.
            card_index  (uint32) Index into controller cards list.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x5c

    pack_uint32(controller, packet, 4)
    pack_uint32(card_index, packet, 8)

    return packet


def put_card_request(controller, card_number, start_date, end_date, door_1, door_2, door_3, door_4, pin):
    '''
    Encodes a put-card request.

        Parameters:
            controller  (uint32)  Controller serial number.
            card_number (uint32)  Card number.
            start_date  (date)    Card 'valid from' date.
            end_date    (date)    Card 'valid until' date.
            door_1      (uint8)   Card access permissions for door 1 (0: none, 1: all, 2-254: time profile ID)
            door_2      (uint8)   Card access permissions for door 2 (0: none, 1: all, 2-254: time profile ID)
            door_3      (uint8)   Card access permissions for door 3 (0: none, 1: all, 2-254: time profile ID)
            door_4      (uint8)   Card access permissions for door 4 (0: none, 1: all, 2-254: time profile ID)
            pin         (uint24)  Card access keypad PIN code

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x50

    pack_uint32(controller, packet, 4)
    pack_uint32(card_number, packet, 8)
    pack_date(start_date, packet, 12)
    pack_date(end_date, packet, 16)
    pack_uint8(door_1, packet, 20)
    pack_uint8(door_2, packet, 21)
    pack_uint8(door_3, packet, 22)
    pack_uint8(door_4, packet, 23)
    pack_pin(pin, packet, 24)

    return packet


def delete_card_request(controller, card_number):
    '''
    Encodes a delete-card request.

        Parameters:
            controller  (uint32)  Controller serial number.
            card_number (uint32)  Card number to delete.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x52

    pack_uint32(controller, packet, 4)
    pack_uint32(card_number, packet, 8)

    return packet


def delete_cards_request(controller):
    '''
    Encodes a delete-cards request.

        Parameters:
            controller  (uint32)  Controller serial number.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x54

    pack_uint32(controller, packet, 4)
    pack_uint32(0x55aaaa55, packet, 8)

    return packet


def get_event_request(controller, event_index):
    '''
    Encodes a get-event request.

        Parameters:
            controller  (uint32)  Controller serial number.
            event_index (uint32)  Index of event in controller events list.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0xb0

    pack_uint32(controller, packet, 4)
    pack_uint32(event_index, packet, 8)

    return packet


def get_event_index_request(controller):
    '''
    Encodes a get-event-index request.

        Parameters:
            controller  (uint32)  Controller serial number.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0xb4

    pack_uint32(controller, packet, 4)

    return packet


def set_event_index_request(controller, event_index):
    '''
    Encodes a set-event-index request.

        Parameters:
            controller  (uint32)  Controller serial number.
            event_index (uint32)  Index from which to start retrieving events.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0xb2

    pack_uint32(controller, packet, 4)
    pack_uint32(event_index, packet, 8)
    pack_uint32(0x55aaaa55, packet, 12)

    return packet


def record_special_events_request(controller, enable):
    '''
    Encodes a record-special-events request.

        Parameters:
            controller  (uint32)  Controller serial number.
            enable      (bool)    Enables/disables door and pushbutton events.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x8e

    pack_uint32(controller, packet, 4)
    pack_bool(enable, packet, 8)

    return packet


def get_time_profile_request(controller, profile_id):
    '''
    Encodes a get-time-profile request.

        Parameters:
            controller  (uint32)  Controller serial number.
            profile_id  (uint8)   Time profile ID [2..254]

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x98

    pack_uint32(controller, packet, 4)
    pack_uint8(profile_id, packet, 8)

    return packet


def set_time_profile_request(controller, profile_id, start_date, end_date, monday, tuesday, wednesday, thursday, friday,
                             saturday, sunday, segment_1_start, segment_1_end, segment_2_start, segment_2_end,
                             segment_3_start, segment_3_end, linked_profile_id):
    '''
    Encodes a set-time-profile request.

        Parameters:
            controller        (uint32)  Controller serial number.
            profile_id        (uint8)   Time profile ID [2..254]
            start_date        (date)    Time profile 'valid from' date.
            end_date          (date)    Time profile 'valid until' date.
            monday            (bool)    Time profile enabled on Monday.
            tuesday           (bool)    Time profile enabled on Tuesday.
            wednesday         (bool)    Time profile enabled on Wednesday.
            thursday          (bool)    Time profile enabled on Thursday.
            friday            (bool)    Time profile enabled on Friday.
            saturday          (bool)    Time profile enabled on Saturday.
            sunday            (bool)    Time profile enabled on Sunday.
            segment_1_start   (time)    Time profile segment 1 start time (HHmm).
            segment_1_end     (time)    Time profile segment 1 end time (HHmm).
            segment_2_start   (time)    Time profile segment 2 start time (HHmm).
            segment_2_end     (time)    Time profile segment 2 end time (HHmm).
            segment_3_start   (time)    Time profile segment 3 start time (HHmm).
            segment_3_end     (time)    Time profile segment 3 end time (HHmm).
            linked_profile_id (uint8)   Next profile ID in chain (0 if none).

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x88

    pack_uint32(controller, packet, 4)
    pack_uint8(profile_id, packet, 8)
    pack_date(start_date, packet, 9)
    pack_date(end_date, packet, 13)
    pack_bool(monday, packet, 17)
    pack_bool(tuesday, packet, 18)
    pack_bool(wednesday, packet, 19)
    pack_bool(thursday, packet, 20)
    pack_bool(friday, packet, 21)
    pack_bool(saturday, packet, 22)
    pack_bool(sunday, packet, 23)
    pack_HHmm(segment_1_start, packet, 24)
    pack_HHmm(segment_1_end, packet, 26)
    pack_HHmm(segment_2_start, packet, 28)
    pack_HHmm(segment_2_end, packet, 30)
    pack_HHmm(segment_3_start, packet, 32)
    pack_HHmm(segment_3_end, packet, 34)
    pack_uint8(linked_profile_id, packet, 36)

    return packet


def delete_all_time_profiles_request(controller):
    '''
    Encodes a delete-all-time-profiles request.

        Parameters:
            controller  (uint32)  Controller serial number.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x8a

    pack_uint32(controller, packet, 4)
    pack_uint32(0x55aaaa55, packet, 8)

    return packet


def add_task_request(controller, start_date, end_date, monday, tuesday, wednesday, thursday, friday, saturday, sunday,
                     start_time, door, task_type, more_cards):
    '''
    Encodes an add-task request.

        Parameters:
            controller  (uint32)  Controller serial number.
            start_date  (datetime)  Task 'valid from' date.
            end_date    (datetime)  Task 'valid until' date.
            monday      (bool)      Task enabled on Monday.
            tuesday     (bool)      Task enabled on Tuesday.
            wednesday   (bool)      Task enabled on Wednesday.
            thursday    (bool)      Task enabled on Thursday.
            friday      (bool)      Task enabled on Friday.
            saturday    (bool)      Task enabled on Saturday.
            sunday      (bool)      Task enabled on Sunday.
            start_time  (time)      Task 'run at' time (HHmm).
            door        (uint8)     Door [1..4] to which task is assigned.
            task_type   (uint8)     Task type
                                    0:  door controlled
                                    1:  door unlocked
                                    2:  door locked
                                    3:  disable time profile
                                    4:  enable time profile
                                    5:  card, no password
                                    6:  card, IN password
                                    7:  card, password
                                    8:  enable 'more cards'
                                    9:  disable 'more cards'
                                    10: trigger once
                                    11: disable pushbutton
                                    12: enable pushbutton
            more_cards  (uint8)     Number of cards for the 'more cards' task.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0xa8

    pack_uint32(controller, packet, 4)
    pack_date(start_date, packet, 8)
    pack_date(end_date, packet, 12)
    pack_bool(monday, packet, 16)
    pack_bool(tuesday, packet, 17)
    pack_bool(wednesday, packet, 18)
    pack_bool(thursday, packet, 19)
    pack_bool(friday, packet, 20)
    pack_bool(saturday, packet, 21)
    pack_bool(sunday, packet, 22)
    pack_HHmm(start_time, packet, 23)
    pack_uint8(door, packet, 25)
    pack_uint8(task_type, packet, 26)
    pack_uint8(more_cards, packet, 27)

    return packet


def refresh_tasklist_request(controller):
    '''
    Encodes a refresh-tasklist request.

        Parameters:
            controller  (uint32)  Controller serial number.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0xac

    pack_uint32(controller, packet, 4)
    pack_uint32(0x55aaaa55, packet, 8)

    return packet


def clear_tasklist_request(controller):
    '''
    Encodes a clear-tasklist request.

        Parameters:
            controller  (uint32)  Controller serial number.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0xa6

    pack_uint32(controller, packet, 4)
    pack_uint32(0x55aaaa55, packet, 8)

    return packet


def set_pc_control_request(controller, enable):
    '''
    Encodes a set-pc-control request.

        Parameters:
            controller  (uint32)  Controller serial number.
            enable      (bool)    Remote access control enabled/disabled.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0xa0

    pack_uint32(controller, packet, 4)
    pack_uint32(0x55aaaa55, packet, 8)
    pack_bool(enable, packet, 12)

    return packet


def set_interlock_request(controller, interlock):
    '''
    Encodes a set-interlock request.

        Parameters:
           controller  (uint32)  Controller serial number (expected to be greater than 0).
           interlock   (uint8)   Door interlock mode:
                                 0:  none
                                 1:  doors 1 and 2 interlocked
                                 2:  doors 2 and 3 interlocked
                                 3:  doors 1 and 2 interlocked, doors 3 and 4 interlocked
                                 4:  doors 1 and 2 and 3 interlocked
                                 8:  doors 1 and 2 and 3 and 4 interlocked

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0xa2

    pack_uint32(controller, packet, 4)
    pack_uint8(interlock, packet, 8)

    return packet


def activate_keypads_request(controller, reader1, reader2, reader3, reader4):
    '''
    Encodes an activate-keypads request.

        Parameters:
           controller (uint32)  Controller serial number (expected to be greater than 0).
           reader1   (bool)    Enables/disable reader 1 access keypad.
           reader2   (bool)    Enables/disable reader 2 access keypad.
           reader3   (bool)    Enables/disable reader 3 access keypad.
           reader4   (bool)    Enables/disable reader 4 access keypad.

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0xa4

    pack_uint32(controller, packet, 4)
    pack_bool(reader1, packet, 8)
    pack_bool(reader2, packet, 9)
    pack_bool(reader3, packet, 10)
    pack_bool(reader4, packet, 11)

    return packet


def set_door_passcodes_request(device_id, door, passcode1, passcode2, passcode3, passcode4):
    '''
    Encodes a set-door-passcodes request.

        Parameters:
           controller (uint32)  Controller serial number (expected to be greater than 0).
           door       (uint8)   Door ID [1..4].
           passcode1  (uint32)  Passcode [0..999999].
           passcode2  (uint32)  Passcode [0..999999].
           passcode3  (uint32)  Passcode [0..999999].
           passcode4  (uint32)  Passcode [0..999999].

        Returns:
            64 byte UDP packet.
    '''
    packet = bytearray(64)

    packet[0] = 0x17
    packet[1] = 0x8c

    pack_uint32(device_id, packet, 4)
    pack_uint8(door, packet, 8)
    pack_uint32(passcode1, packet, 12)
    pack_uint32(passcode2, packet, 16)
    pack_uint32(passcode3, packet, 20)
    pack_uint32(passcode4, packet, 24)

    return packet


def pack_uint8(v, packet, offset):
    '''
    'in-place' packs a uint8 value as a byte into the packet at the offset.

        Parameters:
           v      (uint8)      uint8 value to encode.
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.
    '''
    packet[offset] = v


def pack_uint16(v, packet, offset):
    '''
    'in-place' packs a uint16 value as a 2-byte little endian value into the packet 
    at the offset.

        Parameters:
           v      (uint16)     uint16 value to encode.
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.
    '''
    struct.pack_into('<H', packet, offset, v)


def pack_uint32(v, packet, offset):
    '''
    'in-place' packs a uint16 value as a 4-byte little endian value into the packet 
    at the offset.

        Parameters:
           v      (uint32)     uint32 value to encode.
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.
    '''
    struct.pack_into('<L', packet, offset, v)


def pack_IPv4(v, packet, offset):
    '''
    'in-place' packs an IPv4Address value 4-byte address into the packet at the offset.

        Parameters:
           v      (IPv4Address)  IP address to encode.
           packet (bytearray)    64 byte array.
           offset (int)          Value location in array.
    '''
    packet[offset:offset + 4] = v.packed


def pack_date(v, packet, offset):
    '''
    'in-place' packs a date value as a 4-byte BCD encoded YYYYMMDD value into the packet at the offset.

        Parameters:
           v      (date)       Date value to encode.
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.
    '''
    bcd = f'{v:%Y%m%d}'
    packet[offset:offset + 4] = bytes.fromhex(bcd)


def pack_datetime(v, packet, offset):
    '''
    'in-place' packs a date value as a 7s-byte BCD encoded YYYYMMDDHHmmss value into the packet at 
    the offset.

        Parameters:
           v      (datetime)   Date and time value to encode.
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.
    '''
    bcd = f'{v:%Y%m%d%H%M%S}'
    packet[offset:offset + 7] = bytes.fromhex(bcd)


def pack_HHmm(v, packet, offset):
    '''
    'in-place' packs a short time value as a 2-byte BCD encoded HHmm value into the packet at the offset.

        Parameters:
           v      (time)       Time value to encode. Only the hours and minutes are encoded.
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.
    '''
    bcd = f'{v:%H%M}'
    packet[offset:offset + 2] = bytes.fromhex(bcd)


def pack_bool(v, packet, offset):
    '''
    'in-place' packs a boolean value as a byte value into the packet at the offset. False is encoded
    as 0, True as 1.

        Parameters:
           v      (bool)       Boolean value to encode.
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.
    '''
    packet[offset] = 0x00 if not v else 0x01


def pack_pin(v, packet, offset):
    '''
    'in-place' packs a 24-bit PIN value as a 3-byte little-endian value into the packet at the offset.

        Parameters:
           v      (uint24)     24-bit PIN value to encode.
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.
    '''
    packet[offset] = (v >> 0) & 0x00ff
    packet[offset + 1] = (v >> 8) & 0x0ff
    packet[offset + 2] = (v >> 16) & 0x0ff



'''
UHPPOTE UDP communications wrapper.

Implements the functionality to send and receive 64 byte UDP packets to/from a UHPPOTE 
access controller.
'''

import socket
import struct
import re
import time
import ipaddress

READ_TIMEOUT = struct.pack('ll', 5, 0)  # 5 seconds
WRITE_TIMEOUT = struct.pack('ll', 1, 0)  # 1 second
NO_TIMEOUT = struct.pack('ll', 0, 0)  # (infinite)


class UDP:

    def __init__(self, bind='0.0.0.0', broadcast='255.255.255.255:60000', listen="0.0.0.0:60001", debug=False):
        '''
        Initialises a Uhppote object with the bind address, broadcast address and listen address.

            Parameters:
               bind      (string)  The IPv4 address:port to which to bind when sending a request.
               broadcast (string)  The IPv4 address:port to which to send broadcast UDP messages.
               listen    (string)  The IPv4 address:port on which to listen for events from the
                                   access controllers.
               debug     (bool)    Dumps the sent and received packets to the console if enabled.

            Returns:
               Initialised UDP object.

            Raises:
               Exception  If any of the supplied IPv4 values cannot be translated to a valid IPv4 
                          address:port combination.
        '''
        self._bind = (bind, 0)
        self._broadcast = resolve(broadcast)
        self._listen = resolve(listen)
        self._debug = debug

    def broadcast(self, request):
        '''
        Binds to the bind address from the constructor and then broadcasts a UDP request to the broadcast
        address from the constructor and then waits 5 seconds for the replies from any reponding access 
        controllers.

            Parameters:
               request  (bytearray)  64 byte request packet.

            Returns:
               List of received response packets (may be empty).

            Raises:
               Error  For any socket related errors.
        '''
        self.dump(request)

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)

        try:
            sock.bind(self._bind)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDTIMEO, WRITE_TIMEOUT)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVTIMEO, READ_TIMEOUT)

            sock.sendto(request, self._broadcast)

            return read_all(sock, self._debug)
        finally:
            sock.close()

    def send(self, request):
        '''
        Binds to the bind address from the constructor and then broadcasts a UDP request to the broadcast,
        and then waits 5 seconds for a reply from the destination access controllers.

            Parameters:
               request  (bytearray)  64 byte request packet.

            Returns:
               Received response packet (if any) or None (for set-ip request).

            Raises:
               Error  For any socket related errors.
        '''
        self.dump(request)

        # sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM | socket.SOCK_NONBLOCK)
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)

        try:
            sock.bind(self._bind)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDTIMEO, WRITE_TIMEOUT)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVTIMEO, READ_TIMEOUT)

            sock.sendto(request, self._broadcast)

            if request[1] == 0x96:
                return None

            return read(sock, self._debug)
        finally:
            sock.close()

    def listen(self, onEvent):
        '''
        Binds to the listen address from the constructor and invokes the events handler for
        any received 64 byte UDP packets. Invalid'ish packets are silently discarded.

            Parameters:
               onEvent  (function)  Handler function for received events, with a function signature 
                                    f(packet).

            Returns:
               None.

            Raises:
               Error  For any socket related errors.
        '''
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)

        try:
            sock.bind(self._listen)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVTIMEO, NO_TIMEOUT)

            while True:
                message = sock.recv(1024)
                if len(message) == 64:
                    if self._debug:
                        dump(message)
                    onEvent(message)
        finally:
            sock.close()

    def dump(self, packet):
        '''
        Prints a packet to the console as a formatted hexadecimal string if debug was enabled in the
        constructor.

            Parameters:
               packet  (bytearray)  64 byte UDP packet.

            Returns:
               None.
        '''
        if self._debug:
            dump(packet)


# TODO convert to asyncio
def read(sock, debug):
    '''
    Waits 2.5 seconds for a single 64 byte packet to be received on the socket. Prints the packet to the console
    if debug is True.

        Parameters:
            sock  (socket)  Initialised and open UDP socket.
            debug (bool)    Enables dumping the received packet to the console.

        Returns:
            Received 64 byte UDP packet (or None).
    '''
    sock.settimeout(2.5)

    while True:
        reply = sock.recv(1024)
        if len(reply) == 64:
            if debug:
                dump(reply)
            return reply

    return None


# TODO convert to asyncio
def read_all(sock, debug):
    '''
    Accumulates received 64 byte UDP packets, waiting up to 2.5 seconds for an incoming packet. Prints the 
    packet to the console if debug is True.

        Parameters:
            sock  (socket)  Initialised and open UDP socket.
            debug (bool)    Enables dumping the received packet to the console.

        Returns:
            List of received 64 byte UDP packets (may be empty).
    '''
    sock.settimeout(2.5)

    replies = []
    while True:
        try:
            reply = sock.recv(1024)
            if len(reply) == 64:
                replies.append(reply)
                if debug:
                    dump(reply)
        except socket.timeout:
            break

    return replies


def resolve(addr):
    '''
    Resolves an address:port string into the equivalent ( address, port ) tuple. An addr value
    without a :port suffix defaults to port 60000.

        Parameters:
            addr  (string)  address:port string

        Returns:
            (address, port) as a (string, uint16) tuple
    '''
    match = re.match(r'(.*?):([0-9]+)', addr)
    if match:
        return (match.group(1), int(match.group(2)))
    else:
        address = ipaddress.IPv4Address(addr)
        return (str(address), 60000)


def dump(packet):
    '''
    Prints a packet to the console as a formatted hexadecimal string.

        Parameters:
           packet  (bytearray)  64 byte UDP packet.

        Returns:
            None.
    '''
    for i in range(0, 4):
        offset = i * 16
        u = packet[offset:offset + 8]
        v = packet[offset + 8:offset + 16]

        p = f'{u[0]:02x} {u[1]:02x} {u[2]:02x} {u[3]:02x} {u[4]:02x} {u[5]:02x} {u[6]:02x} {u[7]:02x}'
        q = f'{v[0]:02x} {v[1]:02x} {v[2]:02x} {v[3]:02x} {v[4]:02x} {v[5]:02x} {v[6]:02x} {v[7]:02x}'

        print(f'   {offset:08x}  {p}  {q}')

    print()



'''
Implements a Python wrapper around the UHPPOTE TCP/IP access controller API.
'''

from . import encode
from . import decode
from . import udp


class Uhppote:

    def __init__(self, bind='0.0.0.0', broadcast='255.255.255.255:60000', listen="0.0.0.0:60001", debug=False):
        '''
        Initialises a Uhppote object with the bind address, broadcast address and listen address.

            Parameters:
               bind      (string)  The IPv4 address to which to bind when sending a request.
               broadcast (string)  The IPv4 address:port to which to send broadcast UDP messages.
               listen    (string)  The IPv4 address:port on which to listen for events from the
                                   access controllers.
               debug     (bool)    Enables verbose debugging information.

            Returns:
               Initialised Uhppote object.

            Raises:
               ValueError  If any of the supplied IPv4 values cannot be translated to a valid IPv4 
                           address:port combination.
        '''
        self._udp = udp.UDP(bind, broadcast, listen, debug)

    def get_all_controllers(self):
        '''
        Retrieves a list of all controllers accessible on the local LAN segment.

            Returns:
               []get_controller_response  List of get_controller_responses from access controllers 
                                          on the local LAN segment.

            Raises:
               Exception  If any of the responses from the access controllers cannot be decoded.
        '''
        request = encode.get_controller_request(0)
        replies = self._udp.broadcast(request)

        list = []
        for reply in replies:
            list.append(decode.get_controller_response(reply))

        return list

    def get_controller(self, controller):
        '''
        Retrieves the controller information for an access controller.

            Parameters:
               controller (uint32)  Controller serial number (expected to be greater than 0).

            Returns:
               get_controller_response  Response from access controller to the get-controller request.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.get_controller_request(controller)
        reply = self._udp.send(request)

        if reply != None:
            return decode.get_controller_response(reply)

        return None

    def set_ip(self, controller, address, netmask, gateway):
        '''
        Sets the controller IPv4 address, netmask and gateway address.

            Parameters:
               controller (uint32)       Controller serial number (expected to be greater than 0).
               address    (IPv4Address)  Controller IPv4 address.
               netmask    (IPv4Address)  Controller IPv4 subnet mask.
               gateway    (IPv4Address)  Controller IPv4 gateway address.

            Returns:
               True  For (probably) internal reasons the access controller does not respond to this command.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.set_ip_request(controller, address, netmask, gateway)
        self._udp.send(request)

        return True

    def get_time(self, controller):
        '''
        Retrieves the access controller current date/time.

            Parameters:
               controller (uint32)  Controller serial number (expected to be greater than 0).

            Returns:
               get_time_response  Controller current date/time.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.get_time_request(controller)
        reply = self._udp.send(request)

        if reply != None:
            return decode.get_time_response(reply)

        return None

    def set_time(self, controller, datetime):
        '''
        Sets the access controller current date/time.

            Parameters:
               controller (uint32)   Controller serial number (expected to be greater than 0).
               datetime   (dateime)  Date/time to set.

            Returns:
               set_time_response  Controller current date/time.

            Raises:
               Exception  If the datetime format cannot be encoded or the response from the 
                          access controller cannot be decoded.
        '''
        request = encode.set_time_request(controller, datetime)
        reply = self._udp.send(request)

        if reply != None:
            return decode.set_time_response(reply)

        return None

    def get_status(self, controller):
        '''
        Retrieves the current status of an access controller.

            Parameters:
               controller (uint32)  Controller serial number (expected to be greater than 0).

            Returns:
               get_status_response  Current controller status.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.get_status_request(controller)
        reply = self._udp.send(request)

        if reply != None:
            return decode.get_status_response(reply)

        return None

    def get_listener(self, controller):
        '''
        Retrieves the configured event listener address:port from an access controller.

            Parameters:
               controller (uint32)  Controller serial number (expected to be greater than 0).

            Returns:
               get_listener_response  Current controller event listener UDP address and port.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.get_listener_request(controller)
        reply = self._udp.send(request)

        if reply != None:
            return decode.get_listener_response(reply)

        return None

    def set_listener(self, controller, address, port):
        '''
        Sets an access controller event listener IPv4 address and port.

            Parameters:
               controller (uint32)       Controller serial number (expected to be greater than 0).
               address    (IPv4Address)  IPv4 address of event listener.
               port       (uint16)       UDP port of event listener.

            Returns:
               set_listener_response  Success/fail response from controller.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.set_listener_request(controller, address, port)
        reply = self._udp.send(request)

        if reply != None:
            return decode.set_listener_response(reply)

        return None

    def get_door_control(self, controller, door):
        '''
        Gets the door delay and control mode for an access controller door.

            Parameters:
               controller (uint32)  Controller serial number (expected to be greater than 0).
               door       (uint8)   Door [1..4]

            Returns:
               get_door_control_response  Door delay and control mode.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.get_door_control_request(controller, door)
        reply = self._udp.send(request)

        if reply != None:
            return decode.get_door_control_response(reply)

        return None

    def set_door_control(self, controller, door, mode, delay):
        '''
        Sets the door delay and control mode for an access controller door.

            Parameters:
               controller (uint32)  Controller serial number (expected to be greater than 0).
               door       (uint8)   Door [1..4]
               mode       (uint8)   Control mode (1: normally open, 2: normally closed, 3: controlled)
               delay      (uint8)   Door unlock duration (seconds)

            Returns:
               set_door_control_response  Door delay and control mode.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.set_door_control_request(controller, door, mode, delay)
        reply = self._udp.send(request)

        if reply != None:
            return decode.set_door_control_response(reply)

        return None

    def open_door(self, controller, door):
        '''
        Remotely opens a door controlled by an access controller.

            Parameters:
               controller (uint32)  Controller serial number (expected to be greater than 0).
               door       (uint8)   Door [1..4]

            Returns:
               open_door_response  Door open success/fail response.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.open_door_request(controller, door)
        reply = self._udp.send(request)

        if reply != None:
            return decode.open_door_response(reply)

        return None

    def get_cards(self, controller):
        '''
        Retrieves the number of cards stored in the access controller.

            Parameters:
               controller (uint32)  Controller serial number (expected to be greater than 0).

            Returns:
               get_cards_response  Number of cards stored locally in controller.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.get_cards_request(controller)
        reply = self._udp.send(request)

        if reply != None:
            return decode.get_cards_response(reply)

        return None

    def get_card(self, controller, card_number):
        '''
        Retrieves the card access record for a card number from the access controller.
            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).
               card_number (uint32)  Access card number.

            Returns:
               get_card_response  Card information associated with the card number.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.get_card_request(controller, card_number)
        reply = self._udp.send(request)

        if reply != None:
            return decode.get_card_response(reply)

        return None

    def get_card_by_index(self, controller, card_index):
        '''
        Retrieves the card access record for a card record from the access controller.
            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).
               index       (uint32)  Controller card list record number.

            Returns:
               get_card_by_index_response  Card information associated with the card number.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.get_card_by_index_request(controller, card_index)
        reply = self._udp.send(request)

        if reply != None:
            return decode.get_card_by_index_response(reply)

        return None

    def put_card(self, controller, card_number, start_date, end_date, door_1, door_2, door_3, door_4, pin):
        '''
        Adds (or updates) a card record stored on the access controller.
            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).
               card_number (uint32)  Access card number.
               start_date  (date)    Card 'valid from' date (YYYYMMDD).
               end_date    (date)    Card 'valid until' date (YYYYMMDD).
               door_1      (uint8)   Card access permissions for door 1 (0: none, 1: all, 2-254: time profile ID)
               door_2      (uint8)   Card access permissions for door 2 (0: none, 1: all, 2-254: time profile ID)
               door_3      (uint8)   Card access permissions for door 3 (0: none, 1: all, 2-254: time profile ID)
               door_4      (uint8)   Card access permissions for door 4 (0: none, 1: all, 2-254: time profile ID)
               pin         (uint24)  Card access keypad PIN code (0 for none)

            Returns:
               put_card_response  Card record add/update success/fail.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.put_card_request(controller, card_number, start_date, end_date, door_1, door_2, door_3, door_4,
                                          pin)
        reply = self._udp.send(request)

        if reply != None:
            return decode.put_card_response(reply)

        return None

    def delete_card(self, controller, card_number):
        '''
        Deletes the card record from the access controller.
            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).
               card_number (uint32)  Access card number to delete.

            Returns:
               delete_card_response  Card record delete success/fail.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.delete_card_request(controller, card_number)
        reply = self._udp.send(request)

        if reply != None:
            return decode.delete_card_response(reply)

        return None

    def delete_all_cards(self, controller):
        '''
        Deletes all card records stored on the access controller.
            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).

            Returns:
               delete_all_cards_response  Clear card records success/fail.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.delete_cards_request(controller)
        reply = self._udp.send(request)

        if reply != None:
            return decode.delete_all_cards_response(reply)

        return None

    def get_event(self, controller, event_index):
        '''
        Retrieves a stored event from the access controller.
            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).
               event_index (uint32)  Index of event in controller list.

            Returns:
               get_event_response  Event information.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.get_event_request(controller, event_index)
        reply = self._udp.send(request)

        if reply != None:
            return decode.get_event_response(reply)

        return None

    def get_event_index(self, controller):
        '''
        Retrieves the 'last downloaded event' index from the controller. The downloaded event index
        is a single utility register on the controller that is managed by an application (not by the
        controller).

            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).

            Returns:
               get_event_index_response  Current value of downloaded event index.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.get_event_index_request(controller)
        reply = self._udp.send(request)

        if reply != None:
            return decode.get_event_index_response(reply)

        return None

    def set_event_index(self, controller, event_index):
        '''
        Sets the 'last downloaded event' index on the controller. The downloaded event index is a 
        single utility register on the controller that is managed by an application (not by the
        controller).

            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).
               event_index (uitn32)  Event index to which to set the 'downloaded event' index.

            Returns:
               set_event_index_response  Set event index success/fail response.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.set_event_index_request(controller, event_index)
        reply = self._udp.send(request)

        if reply != None:
            return decode.set_event_index_response(reply)

        return None

    def record_special_events(self, controller, enable):
        '''
        Enables or disables door open and close and pushbutton press events.

            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).
               enable      (bool)    Includes door open and close and pushbutton events in the
                                     events stored and broadcast by the controller.

            Returns:
               record_special_events_response  Record special events success/fail response.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.record_special_events_request(controller, enable)
        reply = self._udp.send(request)

        if reply != None:
            return decode.record_special_events_response(reply)

        return None

    def get_time_profile(self, controller, profile_id):
        '''
        Retrieves a time profile from an access conntroller.

            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).
               profile_id  (uint8)   Time profile ID [2..254] to retrieve.

            Returns:
               get_time_profile_response  Time profile information for the profile ID.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.get_time_profile_request(controller, profile_id)
        reply = self._udp.send(request)

        if reply != None:
            return decode.get_time_profile_response(reply)

        return None

    def set_time_profile(self, controller, profile_id, start_date, end_date, monday, tuesday, wednesday, thursday,
                         friday, saturday, sunday, segment_1_start, segment_1_end, segment_2_start, segment_2_end,
                         segment_3_start, segment_3_end, linked_profile_id):
        '''
        Creates (or updates) a time profile on an access conntroller.

            Parameters:
               controller        (uint32)  Controller serial number (expected to be greater than 0).
               profile_id        (uint8)   Time profile ID [2..254] to retrieve.
               start_date        (date)    Time profile 'valid from' date.
               end_date          (date)    Time profile 'valid until' date.
               monday            (bool)    Time profile enabled on Monday.
               tuesday           (bool)    Time profile enabled on Tuesday.
               wednesday         (bool)    Time profile enabled on Wednesday.
               thursday          (bool)    Time profile enabled on Thursday.
               friday            (bool)    Time profile enabled on Friday.
               saturday          (bool)    Time profile enabled on Saturday.
               sunday            (bool)    Time profile enabled on Sunday.
               segment_1_start   (time)    Time profile segment 1 start time (HHmm).
               segment_1_end     (time)    Time profile segment 1 end time (HHmm).
               segment_2_start   (time)    Time profile segment 2 start time (HHmm).
               segment_2_end     (time)    Time profile segment 2 end time (HHmm).
               segment_3_start   (time)    Time profile segment 3 start time (HHmm).
               segment_3_end     (time)    Time profile segment 3 end time (HHmm).
               linked_profile_id (uint8)   Next profile ID in chain (0 if none).
 
            Returns:
               set_time_profile_response  Set time profile success/fail response.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.set_time_profile_request(controller, profile_id, start_date, end_date, monday, tuesday,
                                                  wednesday, thursday, friday, saturday, sunday, segment_1_start,
                                                  segment_1_end, segment_2_start, segment_2_end, segment_3_start,
                                                  segment_3_end, linked_profile_id)
        reply = self._udp.send(request)

        if reply != None:
            return decode.set_time_profile_response(reply)

        return None

    def delete_all_time_profiles(self, controller):
        '''
        Clears all time profiles from an access conntroller.

            Parameters:
               controller (uint32)  Controller serial number (expected to be greater than 0).

            Returns:
               delete_all_time_profiles_response  Clear time profiles success/fail response.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.delete_all_time_profiles_request(controller)
        reply = self._udp.send(request)

        if reply != None:
            return decode.delete_all_time_profiles_response(reply)

        return None

    def add_task(self, controller, start_date, end_date, monday, tuesday, wednesday, thursday, friday, saturday, sunday,
                 start_time, door, task_type, more_cards):
        '''
        Creates a scheduled task on an access conntroller.

            Parameters:
               controller  (uint32)    Controller serial number (expected to be greater than 0).
               start_date  (datetime)  Task 'valid from' date.
               end_date    (datetime)  Task 'valid until' date.
               monday      (bool)      Task enabled on Monday.
               tuesday     (bool)      Task enabled on Tuesday.
               wednesday   (bool)      Task enabled on Wednesday.
               thursday    (bool)      Task enabled on Thursday.
               friday      (bool)      Task enabled on Friday.
               saturday    (bool)      Task enabled on Saturday.
               sunday      (bool)      Task enabled on Sunday.
               start_time  (time)      Task 'run at' time (HHmm).
               door        (uint8)     Door [1..4] to which task is assigned.
               task_type   (uint8)     Task type
                                       0:  door controlled
                                       1:  door unlocked
                                       2:  door locked
                                       3:  disable time profile
                                       4:  enable time profile
                                       5:  card, no password
                                       6:  card, IN password
                                       7:  card, password
                                       8:  enable 'more cards'
                                       9:  disable 'more cards'
                                       10: trigger once
                                       11: disable pushbutton
                                       12: enable pushbutton
               more_cards  (uint8)     Number of cards for the 'more cards' task.

            Returns:
               add_task_response  Add task success/fail response.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.add_task_request(controller, start_date, end_date, monday, tuesday, wednesday, thursday,
                                          friday, saturday, sunday, start_time, door, task_type, more_cards)
        reply = self._udp.send(request)

        if reply != None:
            return decode.add_task_response(reply)

        return None

    def refresh_tasklist(self, controller):
        '''
        Updates the active tasklist to include tasks added by add_task.

            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).

            Returns:
               refresh_tasklist_response  Refresh tasklist success/fail response.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.refresh_tasklist_request(controller)
        reply = self._udp.send(request)

        if reply != None:
            return decode.refresh_tasklist_response(reply)

        return None

    def clear_tasklist(self, controller):
        '''
        Clears all active and pending tasks.

            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).

            Returns:
               clear_tasklist_response  Clear tasklist success/fail response.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.clear_tasklist_request(controller)
        reply = self._udp.send(request)

        if reply != None:
            return decode.clear_tasklist_response(reply)

        return None

    def set_pc_control(self, controller, enable):
        '''
        Defers access control decisions to a remote host. The remote host is expected to 
        interact with the controller at least once every 30 seconds (typically by enabling
        set_pc_control), failing which the access controller will fallback to the internal
        access control list.

            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).
               enable      (bool)    Enables remote control of access.

            Returns:
               set_pc_control_response  Enable PC control success/fail response.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.set_pc_control_request(controller, enable)
        reply = self._udp.send(request)

        if reply != None:
            return decode.set_pc_control_response(reply)

        return None

    def set_interlock(self, controller, interlock):
        '''
        Sets the door interlock mode for an access controller.

            Parameters:
               controller  (uint32)  Controller serial number (expected to be greater than 0).
               interlock   (uint8)   Door interlock mode:
                                     0:  none
                                     1:  doors 1 and 2 interlocked
                                     2:  doors 2 and 3 interlocked
                                     3:  doors 1 and 2 interlocked, doors 3 and 4 interlocked
                                     4:  doors 1 and 2 and 3 interlocked
                                     8:  doors 1 and 2 and 3 and 4 interlocked

            Returns:
               set_interlock_response  Set interlock success/fail response.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.set_interlock_request(controller, interlock)
        reply = self._udp.send(request)

        if reply != None:
            return decode.set_interlock_response(reply)

        return None

    def activate_keypads(self, controller, reader1, reader2, reader3, reader4):
        '''
        Enables (or disables) the keypad associated with an access reader.

            Parameters:
               controller (uint32)  Controller serial number (expected to be greater than 0).
               reader1    (bool)    Enables/disable reader 1 access keypad
               reader2    (bool)    Enables/disable reader 2 access keypad
               reader3    (bool)    Enables/disable reader 3 access keypad
               reader4    (bool)    Enables/disable reader 4 access keypad

            Returns:
               activate_keypads_response  Activate keypads success/fail response.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.activate_keypads_request(controller, reader1, reader2, reader3, reader4)
        reply = self._udp.send(request)

        if reply != None:
            return decode.activate_keypads_response(reply)

        return None

    def set_door_passcodes(self, controller, door, passcode1, passcode2, passcode3, passcode4):
        '''
        Sets up to four supervisor passcodes for a door. The passcodes override any other access 
        restrictions and a valid passcode is in the range [0..999999], with 0 corresponding to 
        'no code'.

            Parameters:
               controller (uint32)  Controller serial number (expected to be greater than 0).
               door       (uint8)   Door ID [1..4].
               passcode1  (uint32)  Passcode [0..999999].
               passcode2  (uint32)  Passcode [0..999999].
               passcode3  (uint32)  Passcode [0..999999].
               passcode4  (uint32)  Passcode [0..999999].

            Returns:
               set_door_passcodes  Set door passcodes success/fail response.

            Raises:
               Exception  If the response from the access controller cannot be decoded.
        '''
        request = encode.set_door_passcodes_request(controller, door, passcode1, passcode2, passcode3, passcode4)
        reply = self._udp.send(request)

        if reply != None:
            return decode.set_door_passcodes_response(reply)

        return None

    def listen(self, onEvent):
        '''
        Establishes a listener for events from the access controllers by binding to the UDP listen 
        address from the constructor.

            Parameters:
               onEvent  (function)  Handler function for received events, with a function signature 
                                    f(event).

            Returns:
               None
        '''

        def handler(packet):
            try:
                onEvent(decode.event(packet))
            except BaseException as err:
                print('   *** ERROR {}'.format(err))

        self._udp.listen(lambda packet: handler(packet))

        return None



'''
UHPPOTE request packet decoder.

Decodes a UHPPOTE access controller 64 byte UDP response packet into the
equivalent Python object.
'''

import datetime
import struct

from ipaddress import IPv4Address
from dataclasses import dataclass

from .structs import GetControllerResponse
from .structs import GetTimeResponse
from .structs import SetTimeResponse
from .structs import GetStatusResponse
from .structs import GetListenerResponse
from .structs import SetListenerResponse
from .structs import GetDoorControlResponse
from .structs import SetDoorControlResponse
from .structs import OpenDoorResponse
from .structs import GetCardsResponse
from .structs import GetCardResponse
from .structs import GetCardByIndexResponse
from .structs import PutCardResponse
from .structs import DeleteCardResponse
from .structs import DeleteAllCardsResponse
from .structs import GetEventResponse
from .structs import GetEventIndexResponse
from .structs import SetEventIndexResponse
from .structs import RecordSpecialEventsResponse
from .structs import GetTimeProfileResponse
from .structs import SetTimeProfileResponse
from .structs import DeleteAllTimeProfilesResponse
from .structs import AddTaskResponse
from .structs import RefreshTasklistResponse
from .structs import ClearTasklistResponse
from .structs import SetPcControlResponse
from .structs import SetInterlockResponse
from .structs import ActivateKeypadsResponse
from .structs import SetDoorPasscodesResponse
from .structs import Event
from .structs import PIN


def get_controller_response(packet):
    '''
    Decodes a get-controller response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            GetControllerResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x94:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return GetControllerResponse(
        unpack_uint32(packet, 4),
        unpack_ipv4(packet, 8),
        unpack_ipv4(packet, 12),
        unpack_ipv4(packet, 16),
        unpack_mac(packet, 20),
        unpack_version(packet, 26),
        unpack_date(packet, 28),
    )


def get_time_response(packet):
    '''
    Decodes a get-time response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            GetTimeResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x32:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return GetTimeResponse(
        unpack_uint32(packet, 4),
        unpack_datetime(packet, 8),
    )


def set_time_response(packet):
    '''
    Decodes a set-time response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            SetTimeResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x30:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return SetTimeResponse(
        unpack_uint32(packet, 4),
        unpack_datetime(packet, 8),
    )


def get_status_response(packet):
    '''
    Decodes a get-status response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            GetStatusResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x20:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    event_index = unpack_uint32(packet, 8)

    # no event in response ?
    if event_index == 0:
        return GetStatusResponse(
            unpack_uint32(packet, 4),
            unpack_shortdate(packet, 51),
            unpack_time(packet, 37),
            unpack_bool(packet, 28),
            unpack_bool(packet, 29),
            unpack_bool(packet, 30),
            unpack_bool(packet, 31),
            unpack_bool(packet, 32),
            unpack_bool(packet, 33),
            unpack_bool(packet, 34),
            unpack_bool(packet, 35),
            unpack_uint8(packet, 49),
            unpack_uint8(packet, 50),
            unpack_uint8(packet, 36),
            unpack_uint8(packet, 48),
            unpack_uint32(packet, 8),
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            unpack_uint32(packet, 40),
        )
    else:
        return GetStatusResponse(
            unpack_uint32(packet, 4),
            unpack_shortdate(packet, 51),
            unpack_time(packet, 37),
            unpack_bool(packet, 28),
            unpack_bool(packet, 29),
            unpack_bool(packet, 30),
            unpack_bool(packet, 31),
            unpack_bool(packet, 32),
            unpack_bool(packet, 33),
            unpack_bool(packet, 34),
            unpack_bool(packet, 35),
            unpack_uint8(packet, 49),
            unpack_uint8(packet, 50),
            unpack_uint8(packet, 36),
            unpack_uint8(packet, 48),
            unpack_uint32(packet, 8),
            unpack_uint8(packet, 12),
            unpack_bool(packet, 13),
            unpack_uint8(packet, 14),
            unpack_uint8(packet, 15),
            unpack_uint32(packet, 16),
            unpack_optional_datetime(packet, 20),
            unpack_uint8(packet, 27),
            unpack_uint32(packet, 40),
        )


def get_listener_response(packet):
    '''
    Decodes a get-listener response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            GetListenerResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x92:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return GetListenerResponse(
        unpack_uint32(packet, 4),
        unpack_ipv4(packet, 8),
        unpack_uint16(packet, 12),
    )


def set_listener_response(packet):
    '''
    Decodes a set-listener response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            SetListenerResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x90:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return SetListenerResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def get_door_control_response(packet):
    '''
    Decodes a get-door-control response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            GetDoorControlResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x82:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return GetDoorControlResponse(
        unpack_uint32(packet, 4),
        unpack_uint8(packet, 8),
        unpack_uint8(packet, 9),
        unpack_uint8(packet, 10),
    )


def set_door_control_response(packet):
    '''
    Decodes a set-door-control response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            SetDoorControlResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x80:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return SetDoorControlResponse(
        unpack_uint32(packet, 4),
        unpack_uint8(packet, 8),
        unpack_uint8(packet, 9),
        unpack_uint8(packet, 10),
    )


def open_door_response(packet):
    '''
    Decodes an open-door response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            OpenDoorResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x40:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return OpenDoorResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def get_cards_response(packet):
    '''
    Decodes a get-cards response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            GetCardsResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x58:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return GetCardsResponse(
        unpack_uint32(packet, 4),
        unpack_uint32(packet, 8),
    )


def get_card_response(packet):
    '''
    Decodes a get-card response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            GetCardResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x5a:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return GetCardResponse(
        unpack_uint32(packet, 4),
        unpack_uint32(packet, 8),
        unpack_optional_date(packet, 12),
        unpack_optional_date(packet, 16),
        unpack_uint8(packet, 20),
        unpack_uint8(packet, 21),
        unpack_uint8(packet, 22),
        unpack_uint8(packet, 23),
        unpack_pin(packet, 24),
    )


def get_card_by_index_response(packet):
    '''
    Decodes a get-card-by-index response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            GetCardByIndexResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x5c:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return GetCardByIndexResponse(
        unpack_uint32(packet, 4),
        unpack_uint32(packet, 8),
        unpack_optional_date(packet, 12),
        unpack_optional_date(packet, 16),
        unpack_uint8(packet, 20),
        unpack_uint8(packet, 21),
        unpack_uint8(packet, 22),
        unpack_uint8(packet, 23),
        unpack_pin(packet, 24),
    )


def put_card_response(packet):
    '''
    Decodes a put-card response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            PutCardResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x50:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return PutCardResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def delete_card_response(packet):
    '''
    Decodes a delete-card response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            DeleteCardResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x52:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return DeleteCardResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def delete_all_cards_response(packet):
    '''
    Decodes a delete-all-cards response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            DeleteAllCardsResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x54:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return DeleteAllCardsResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def get_event_response(packet):
    '''
    Decodes a get-event response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            GetEventResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0xb0:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return GetEventResponse(
        unpack_uint32(packet, 4),
        unpack_uint32(packet, 8),
        unpack_uint8(packet, 12),
        unpack_bool(packet, 13),
        unpack_uint8(packet, 14),
        unpack_uint8(packet, 15),
        unpack_uint32(packet, 16),
        unpack_optional_datetime(packet, 20),
        unpack_uint8(packet, 27),
    )


def get_event_index_response(packet):
    '''
    Decodes a get-event-index response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            GetEventIndexResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0xb4:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return GetEventIndexResponse(
        unpack_uint32(packet, 4),
        unpack_uint32(packet, 8),
    )


def set_event_index_response(packet):
    '''
    Decodes a set-event-index response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            SetEventIndexResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0xb2:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return SetEventIndexResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def record_special_events_response(packet):
    '''
    Decodes a record-special-events response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            RecordSpecialEventsResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x8e:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return RecordSpecialEventsResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def get_time_profile_response(packet):
    '''
    Decodes a get-time-profile response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            GetTimeProfileResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x98:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return GetTimeProfileResponse(
        unpack_uint32(packet, 4),
        unpack_uint8(packet, 8),
        unpack_optional_date(packet, 9),
        unpack_optional_date(packet, 13),
        unpack_bool(packet, 17),
        unpack_bool(packet, 18),
        unpack_bool(packet, 19),
        unpack_bool(packet, 20),
        unpack_bool(packet, 21),
        unpack_bool(packet, 22),
        unpack_bool(packet, 23),
        unpack_hhmm(packet, 24),
        unpack_hhmm(packet, 26),
        unpack_hhmm(packet, 28),
        unpack_hhmm(packet, 30),
        unpack_hhmm(packet, 32),
        unpack_hhmm(packet, 34),
        unpack_uint8(packet, 36),
    )


def set_time_profile_response(packet):
    '''
    Decodes a set-time-profile response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            SetTimeProfileResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x88:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return SetTimeProfileResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def delete_all_time_profiles_response(packet):
    '''
    Decodes a delete-all-time-profiles response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            DeleteAllTimeProfilesResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x8a:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return DeleteAllTimeProfilesResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def add_task_response(packet):
    '''
    Decodes an add-task response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            AddTaskResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0xa8:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return AddTaskResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def refresh_tasklist_response(packet):
    '''
    Decodes a refresh-tasklist response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            RefreshTasklistResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0xac:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return RefreshTasklistResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def clear_tasklist_response(packet):
    '''
    Decodes a clear-tasklist response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            ClearTasklistResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0xa6:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return ClearTasklistResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def set_pc_control_response(packet):
    '''
    Decodes a set-pc-control response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            SetPcControlResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0xa0:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return SetPcControlResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def set_interlock_response(packet):
    '''
    Decodes a set-interlock response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            SetInterlockResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0xa2:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return SetInterlockResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def activate_keypads_response(packet):
    '''
    Decodes an activate-keypads response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            ActivateKeypadsResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0xa4:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return ActivateKeypadsResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def set_door_passcodes_response(packet):
    '''
    Decodes a set-door-passcodes response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            SetDoorPasscodesResponse initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x8c:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    return SetDoorPasscodesResponse(
        unpack_uint32(packet, 4),
        unpack_bool(packet, 8),
    )


def event(packet):
    '''
    Decodes an event packet response.

        Parameters:
            packet  (bytearray)  64 byte UDP packet.

        Returns:
            Event initialised from the UDP packet.

        Raises:
            ValueError If the packet is not 64 bytes, has an invalid start-of-message byte or has
                       the incorrect message type.
    '''
    if len(packet) != 64:
        raise ValueError(f'invalid reply packet length ({len(packet)})')

    # Ref. v6.62 firmware event
    if packet[0] != 0x17 and (packet[0] != 0x19 or packet[1] != 0x20):
        raise ValueError(f'invalid reply start of message byte ({packet[0]:02x})')

    if packet[1] != 0x20:
        raise ValueError(f'invalid reply function code ({packet[1]:02x})')

    # yapf: disable
    return Event(
        unpack_uint32(packet, 4),       # controller
        unpack_uint32(packet, 8),       # event index
        unpack_uint8(packet, 12),       # event type
        unpack_bool(packet, 13),        # event access granted
        unpack_uint8(packet, 14),       # event door
        unpack_uint8(packet, 15),       # event direction
        unpack_uint32(packet, 16),      # event card number
        unpack_datetime(packet, 20),    # event timestamp
        unpack_uint8(packet, 27),       # event reason
        unpack_shortdate(packet, 51),   # system date
        unpack_time(packet, 37),        # system time
        unpack_bool(packet, 28),        # door 1 open
        unpack_bool(packet, 29),        # door 2 open
        unpack_bool(packet, 30),        # door 3 open
        unpack_bool(packet, 31),        # door 4 open
        unpack_bool(packet, 32),        # door 1 button
        unpack_bool(packet, 33),        # door 2 button
        unpack_bool(packet, 34),        # door 3 button
        unpack_bool(packet, 35),        # door 4 button
        unpack_uint8(packet, 49),       # relays
        unpack_uint8(packet, 50),       # inputs
        unpack_uint8(packet, 36),       # system error
        unpack_uint8(packet, 48),       # special info
        unpack_uint32(packet, 40),      # sequence no
    )
    # yapf: enable


def unpack_uint8(packet, offset):
    '''
    Unpacks the uint8 value from the packet at the offset.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           uint8 value.
    '''
    return packet[offset]


def unpack_uint16(packet, offset):
    '''
    Unpacks the 2-byte little endian uint16 value from the packet at the offset.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           uint16 value.
    '''
    return struct.unpack_from('<H', packet, offset)[0]


def unpack_uint32(packet, offset):
    '''
    Unpacks the 4-byte little endian uint32 value from the packet at the offset.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           uint32 value.
    '''
    return struct.unpack_from('<L', packet, offset)[0]


def unpack_ipv4(packet, offset):
    '''
    Unpacks the 4-byte IP address value from the packet at the offset.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           IPv4Address value.
    '''
    return IPv4Address(f'{packet[offset]}.{packet[offset+1]}.{packet[offset+2]}.{packet[offset+3]}')


def unpack_mac(packet, offset):
    '''
    Unpacks the 6-byte MAC address value from the packet at the offset.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           MAC address as a colon-seperated hexadecimal string.
    '''
    return '{:02x}:{:02x}:{:02x}:{:02x}:{:02x}:{:02x}'.format(*packet[offset:offset + 7])


def unpack_version(packet, offset):
    '''
    Unpacks the 2-byte BCD encoded version value from the packet at the offset.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           Version string.
    '''
    return 'v{:x}.{:02x}'.format(*packet[offset:offset + 2])


def unpack_date(packet, offset):
    '''
    Unpacks the 4-byte BCD encoded YYYYMMDD value from the packet at the offset.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           datetime value (or None if the date/time valid is invalid)
    '''
    bcd = '{:02x}{:02x}{:02x}{:02x}'.format(*packet[offset:offset + 4])

    try:
        return datetime.datetime.strptime(bcd, '%Y%m%d').date()
    except ValueError:
        return None


def unpack_shortdate(packet, offset):
    '''
    Unpacks the 3-byte BCD encoded YYMMDD value from the packet at the offset,
    automatically adding the CC value.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           datetime value (or None if the date/time valid is invalid)
    '''
    bcd = '20{:02x}{:02x}{:02x}'.format(*packet[offset:offset + 3])

    try:
        return datetime.datetime.strptime(bcd, '%Y%m%d').date()
    except ValueError:
        return None


def unpack_optional_date(packet, offset):
    '''
    Unpacks the 4-byte BCD encoded YYYYMMDD value from the packet at the offset.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           datetime value (or None if the date/time valid is invalid)
    '''
    bcd = '{:02x}{:02x}{:02x}{:02x}'.format(*packet[offset:offset + 4])

    try:
        return datetime.datetime.strptime(bcd, '%Y%m%d').date()
    except ValueError:
        return None


def unpack_datetime(packet, offset):
    '''
    Unpacks the 7-byte BCD encoded YYYYMMDDHHmmss value from the packet at the offset.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           datetime value (or None if the BCD value is not a valid date/time).
    '''
    bcd = '{:02x}{:02x}{:02x}{:02x}{:02x}{:02x}{:02x}'.format(*packet[offset:offset + 7])

    try:
        return datetime.datetime.strptime(bcd, '%Y%m%d%H%M%S')
    except ValueError:
        return None


def unpack_optional_datetime(packet, offset):
    '''
    Unpacks the 7-byte BCD encoded YYYYMMDDHHmmss value from the packet at the offset.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           datetime value (or None if the BCD value is not a valid date/time).
    '''
    bcd = '{:02x}{:02x}{:02x}{:02x}{:02x}{:02x}{:02x}'.format(*packet[offset:offset + 7])

    try:
        return datetime.datetime.strptime(bcd, '%Y%m%d%H%M%S')
    except ValueError:
        return None


def unpack_time(packet, offset):
    '''
    Unpacks the 3-byte BCD encoded HHmmss value from the packet at the offset.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           datetime value (or None if the BCD value is not a valid date/time).
    '''
    bcd = '{:02x}{:02x}{:02x}'.format(*packet[offset:offset + 3])

    try:
        return datetime.datetime.strptime(bcd, '%H%M%S').time()
    except ValueError:
        return None


def unpack_hhmm(packet, offset):
    '''
    Unpacks the 2-byte BCD encoded HHmm value from the packet at the offset.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           datetime value (or None if the BCD value is not a valid date/time).
    '''
    bcd = '{:02x}{:02x}'.format(*packet[offset:offset + 2])

    try:
        return datetime.datetime.strptime(bcd, '%H%M').time()
    except ValueError:
        return None


def unpack_bool(packet, offset):
    '''
    Unpacks the byte from the packet at the offset as a boolean value. 0 is translated
    to False and 'not zero' to True. The access controller should only fill this field
    with either a one or a zero.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           datetime value.
    '''
    return packet[offset] != 0x00


def unpack_pin(packet, offset):
    '''
    Unpacks the 3-byte PIN value from the packet at the offset.

        Parameters:
           packet (bytearray)  64 byte array.
           offset (int)        Value location in array.

        Returns:
           PIN  value.
    '''
    v = packet[offset + 2] & 0x0ff
    v <<= 8
    v |= packet[offset + 1] & 0x0ff
    v <<= 8
    v |= packet[offset] & 0x00ff

    return v



'''
Shared dataclass definitions.
'''

import datetime
import struct

from ipaddress import IPv4Address
from dataclasses import dataclass
from typing import NewType

PIN = NewType('PIN', int)


@dataclass
class GetControllerResponse:
    '''
    Container class for the decoded response from a get-controller request.

       Fields:
          controller   (uint32)       Serial number.
          ip_address   (IPv4Address)  IP address.
          subnet_mask  (IPv4Address)  Subnet mask.
          gateway      (IPv4Address)  Gateway IP address.
          mac_address  (string)       MAC address (XX:XX:XX:XX:XX:XX).
          version      (string)       Firmware version (vN.NN).
          date         (date)         Release date (YYYY-MM-DD).
    '''
    controller: int
    ip_address: IPv4Address
    subnet_mask: IPv4Address
    gateway: IPv4Address
    mac_address: str
    version: str
    date: datetime.date


@dataclass
class GetTimeResponse:
    '''
    Container class for the decoded response from a get-time request.

       Fields:
          controller   (uint32)    Controller serial number.
          datetime     (datetime)  Controller system date/time. None if
                                   the returned date/time is invalid.
    '''
    controller: int
    datetime: datetime.datetime


@dataclass
class SetTimeResponse:
    '''
    Container class for the decoded response from a set-time request.

       Fields:
          controller   (uint32)    Controller serial number.
          datetime     (datetime)  Controller system date/time.
    '''
    controller: int
    datetime: datetime.datetime


@dataclass
class GetStatusResponse:
    '''
    Container class for the decoded response from a get-status request.

       Fields:
          controller           (uint32)    Controller serial number.
          system_date          (date)      Controller system date.
          system_time          (time)      Controller system time.
          door_1_open          (bool)      Door 1 locked/unlocked.
          door_2_open          (bool)      Door 2 locked/unlocked.
          door_3_open          (bool)      Door 3 locked/unlocked.
          door_4_open          (bool)      Door 4 locked/unlocked.
          door_1_button        (bool)      Pushbutton 1 pressed/released.
          door_2_button        (bool)      Pushbutton 2 pressed/released.
          door_3_button        (bool)      Pushbutton 3 pressed/released.
          door_4_button        (bool)      Pushbutton 4 pressed/released.
          relays               (uint8)     Bit array of relay states.
          inputs               (uint8)     Bit array of door sensor states.
          system_error         (uint8)     System error code.
          special_info         (uint8)     Absolutely no idea.
          event_index          (uint32)    Index of last recorded event.
          event_type           (uint32)    Type of last recorded event.
          event_access_granted (bool)      Last event access granted/denied.
          event_door           (uin8)      Last event door no. [1..4].
          event_direction      (uint8)     Last event direction (0: in, 1: out).
          event_card           (uint32)    Last event card number.
          event_timestamp      (datetime)  Last event timestamp.
          event_reason         (uint8)     Last event access granted/denied reason code.
          sequence_no          (uint32)    Packet sequence number.
    '''
    controller: int
    system_date: datetime.date
    system_time: datetime.time
    door_1_open: bool
    door_2_open: bool
    door_3_open: bool
    door_4_open: bool
    door_1_button: bool
    door_2_button: bool
    door_3_button: bool
    door_4_button: bool
    relays: int
    inputs: int
    system_error: int
    special_info: int
    event_index: int
    event_type: int
    event_access_granted: bool
    event_door: int
    event_direction: int
    event_card: int
    event_timestamp: datetime.datetime
    event_reason: int
    sequence_no: int


@dataclass
class GetListenerResponse:
    '''
    Container class for the decoded response from a get-listener request.

       Fields:
          controller  (uint32)       Controller serial number.
          address     (IPv4Address)  Configured event listener IP address.
          port        (uint16)       Configured event listener UDP port.
    '''
    controller: int
    address: IPv4Address
    port: int


@dataclass
class SetListenerResponse:
    '''
    Container class for the decoded response from a set-listener request.

       Fields:
          controller  (uint32)  Controller serial number.
          ok          (bool)    Succeeded/failed.
    '''
    controller: int
    ok: bool


@dataclass
class GetDoorControlResponse:
    '''
    Container class for the decoded response from a get-door-control request.

       Fields:
          controller  (uint32)  Controller serial number.
          door        (uint8)   Door no.[1..4]
          mode        (uint8)   Door control mode (1: normally open, 2: normally closed, 3: controlled)
          delay       (uint8)   Door unlock duration (seconds)
    '''
    controller: int
    door: int
    mode: int
    delay: int


@dataclass
class SetDoorControlResponse:
    '''
    Container class for the decoded response from a set-door-control request.

       Fields:
          controller  (uint32)  Controller serial number.
          door        (uint8)   Door no.[1..4]
          mode        (uint8)   Door control mode (1: normally open, 2: normally closed, 3: controlled)
          delay       (uint8)   Door unlock duration (seconds)
    '''
    controller: int
    door: int
    mode: int
    delay: int


@dataclass
class OpenDoorResponse:
    '''
    Container class for the decoded response from an open-door request.

       Fields:
          controller  (uint32)  Controller serial number.
          opened      (bool)    Succeeded/failed.
    '''
    controller: int
    opened: bool


@dataclass
class GetCardsResponse:
    '''
    Container class for the decoded response from an open-door request.

       Fields:
          controller  (uint32)  Controller serial number.
          cards       (uint32)  Number of cards stored on controller.
    '''
    controller: int
    cards: int


@dataclass
class GetCardResponse:
    '''
    Container class for the decoded response from a get-card request.

       Fields:
          controller  (uint32)  Controller serial number.
          card_number (uint32)  Card number.
          start_date  (date)    Card 'valid from' date.
          end_date    (date)    Card 'valid until' date.
          end_date    (date)    Card 'valid until' date.
          door_1      (uint8)   Card access permissions for door 1 (0: none, 1: all, 2-254: time profile ID)
          door_2      (uint8)   Card access permissions for door 2 (0: none, 1: all, 2-254: time profile ID)
          door_3      (uint8)   Card access permissions for door 3 (0: none, 1: all, 2-254: time profile ID)
          door_4      (uint8)   Card access permissions for door 4 (0: none, 1: all, 2-254: time profile ID)
          pin         (uint24)  Card access keypad PIN code (0 for none)
    '''
    controller: int
    card_number: int
    start_date: datetime.date
    end_date: datetime.date
    door_1: int
    door_2: int
    door_3: int
    door_4: int
    pin: PIN


@dataclass
class GetCardByIndexResponse:
    '''
    Container class for the decoded response from a get-card-by-index request.

       Fields:
          controller  (uint32)  Controller serial number.
          card_number (uint32)  Card number.
          start_date  (date)    Card 'valid from' date.
          end_date    (date)    Card 'valid until' date.
          door_1      (uint8)   Card access permissions for door 1 (0: none, 1: all, 2-254: time profile ID)
          door_2      (uint8)   Card access permissions for door 2 (0: none, 1: all, 2-254: time profile ID)
          door_3      (uint8)   Card access permissions for door 3 (0: none, 1: all, 2-254: time profile ID)
          door_4      (uint8)   Card access permissions for door 4 (0: none, 1: all, 2-254: time profile ID)
          pin         (uint24)  Card access keypad PIN code (0 for none)
    '''
    controller: int
    card_number: int
    start_date: datetime.date
    end_date: datetime.date
    door_1: int
    door_2: int
    door_3: int
    door_4: int
    pin: PIN


@dataclass
class PutCardResponse:
    '''
    Container class for the decoded response from an open-door request.

       Fields:
          controller  (uint32)  Controller serial number.
          stored      (bool)    Succeeded/failed.
    '''
    controller: int
    stored: bool


@dataclass
class DeleteCardResponse:
    '''
    Container class for the decoded response from a delete-card request.

       Fields:
          controller  (uint32)  Controller serial number.
          deleted     (bool)    Succeeded/failed.
    '''
    controller: int
    deleted: bool


@dataclass
class DeleteAllCardsResponse:
    '''
    Container class for the decoded response from a delete-all-cards request.

       Fields:
          controller  (uint32)  Controller serial number.
          deleted     (bool)    Succeeded/failed.
    '''
    controller: int
    deleted: bool


@dataclass
class GetEventResponse:
    '''
    Container class for the decoded response from a get-event request.

       Fields:
          controller      (uint32)    Controller serial number.
          index           (uint32)    Index of last recorded event.
          event_type      (uint32)    Type of last recorded event.
          access_granted  (bool)      Last event access granted/denied.
          door            (uin8)      Last event door no. [1..4].
          direction       (uint8)     Last event direction (0: in, 1: out).
          card            (uint32)    Last event card number.
          timestamp       (datetime)  Last event timestamp.
          reason          (uint8)     Last event access granted/denied reason code.
    '''
    controller: int
    index: int
    event_type: int
    access_granted: bool
    door: int
    direction: int
    card: int
    timestamp: datetime.datetime
    reason: int


@dataclass
class GetEventIndexResponse:
    '''
    Container class for the decoded response from a get-event-index request.

       Fields:
          controller      (uint32)    Controller serial number.
          index           (uint32)    Downloaded event index.
    '''
    controller: int
    event_index: int


@dataclass
class SetEventIndexResponse:
    '''
    Container class for the decoded response from a set-event-index request.

       Fields:
          controller  (uint32)  Controller serial number.
          updated     (bool)    Succeeded/failed.
    '''
    controller: int
    updated: bool


@dataclass
class RecordSpecialEventsResponse:
    '''
    Container class for the decoded response from a record-special-events request.

       Fields:
          controller  (uint32)  Controller serial number.
          updated     (bool)    Succeeded/failed.
    '''
    controller: int
    updated: bool


@dataclass
class GetTimeProfileResponse:
    '''
    Container class for the decoded response from a get-time-profile request.

       Fields:
          controller        (uint32)  Controller serial number.
          profile_id        (uint8)   Time profile ID [2..254].
          start_date        (date)    Time profile 'valid from' date.
          end_date          (date)    Time profile 'valid until' date.
          monday            (bool)    Time profile enabled on Monday.
          tuesday           (bool)    Time profile enabled on Tuesday.
          wednesday         (bool)    Time profile enabled on Wednesday.
          thursday          (bool)    Time profile enabled on Thursday.
          friday            (bool)    Time profile enabled on Friday.
          saturday          (bool)    Time profile enabled on Saturday.
          sunday            (bool)    Time profile enabled on Sunday.
          segment_1_start   (time)    Time profile segment 1 start time (HHmm).
          segment_1_end     (time)    Time profile segment 1 end time (HHmm).
          segment_2_start   (time)    Time profile segment 2 start time (HHmm).
          segment_2_end     (time)    Time profile segment 2 end time (HHmm).
          segment_3_start   (time)    Time profile segment 3 start time (HHmm).
          segment_3_end     (time)    Time profile segment 3 end time (HHmm).
          linked_profile_id (uint8)   Next profile ID in chain (0 if none).
    '''
    controller: int
    profile_id: int
    start_date: datetime.date
    end_date: datetime.date
    monday: bool
    tuesday: bool
    wednesday: bool
    thursday: bool
    friday: bool
    saturday: bool
    sunday: bool
    segment_1_start: datetime.time
    segment_1_end: datetime.time
    segment_2_start: datetime.time
    segment_2_end: datetime.time
    segment_3_start: datetime.time
    segment_3_end: datetime.time
    linked_profile_id: int


@dataclass
class SetTimeProfileResponse:
    '''
    Container class for the decoded response from a set-time-profile request.

       Fields:
          controller  (uint32)  Controller serial number.
          stored      (bool)    Succeeded/failed.
    '''
    controller: int
    stored: bool


@dataclass
class DeleteAllTimeProfilesResponse:
    '''
    Container class for the decoded response from a delete-all-time-profiles request.

       Fields:
          controller  (uint32)  Controller serial number.
          deleted     (bool)    Succeeded/failed.
    '''
    controller: int
    deleted: bool


@dataclass
class AddTaskResponse:
    '''
    Container class for the decoded response from an add-task request.

       Fields:
          controller  (uint32)  Controller serial number.
          added       (bool)    Succeeded/failed.
    '''
    controller: int
    added: bool


@dataclass
class RefreshTasklistResponse:
    '''
    Container class for the decoded response from a refersh-tasklist request.

       Fields:
          controller  (uint32)  Controller serial number.
          refreshed   (bool)    Succeeded/failed.
    '''
    controller: int
    refreshed: bool


@dataclass
class ClearTasklistResponse:
    '''
    Container class for the decoded response from a clear-tasklist request.

       Fields:
          controller  (uint32)  Controller serial number.
          cleared     (bool)    Succeeded/failed.
    '''
    controller: int
    cleared: bool


@dataclass
class SetPcControlResponse:
    '''
    Container class for the decoded response from a set-pc-control request.

       Fields:
          controller  (uint32)  Controller serial number.
          ok          (bool)    Succeeded/failed.
    '''
    controller: int
    ok: bool


@dataclass
class SetInterlockResponse:
    '''
    Container class for the decoded response from a set-interlock request.

       Fields:
          controller  (uint32)  Controller serial number.
          ok          (bool)    Succeeded/failed.
    '''
    controller: int
    ok: bool


@dataclass
class ActivateKeypadsResponse:
    '''
    Container class for the decoded response from an activate-keypads request.

       Fields:
          controller  (uint32)  Controller serial number.
          ok          (bool)    Succeeded/failed.
    '''
    controller: int
    ok: bool


@dataclass
class SetDoorPasscodesResponse:
    '''
    Container class for the decoded response from a set-door-passcodes request.

       Fields:
          controller  (uint32)  Controller serial number.
          ok          (bool)    Succeeded/failed.
    '''
    controller: int
    ok: bool


@dataclass
class Event:
    '''
    Container class for a decoded event packet.

       Fields:
          controller           (uint32)    Controller serial number.
          event_index          (uint32)    Index of last recorded event.
          event_type           (uint32)    Type of last recorded event.
          event_access_granted (bool)      Last event access granted/denied.
          event_door           (uin8)      Last event door no. [1..4].
          event_direction      (uint8)     Last event direction (0: in, 1: out).
          event_card           (uint32)    Last event card number.
          event_timestamp      (datetime)  Last event timestamp.
          event_reason         (uint8)     Last event access granted/denied reason code.
          system_date          (date)      Controller system date.
          system_time          (time)      Controller system time.
          door_1_open          (bool)      Door 1 locked/unlocked.
          door_2_open          (bool)      Door 2 locked/unlocked.
          door_3_open          (bool)      Door 3 locked/unlocked.
          door_4_open          (bool)      Door 4 locked/unlocked.
          door_1_button        (bool)      Pushbutton 1 pressed/released.
          door_2_button        (bool)      Pushbutton 2 pressed/released.
          door_3_button        (bool)      Pushbutton 3 pressed/released.
          door_4_button        (bool)      Pushbutton 4 pressed/released.
          relays               (uint8)     Bit array of relay states.
          inputs               (uint8)     Bit array of door sensor states.
          system_error         (uint8)     System error code.
          special_info         (uint8)     Absolutely no idea.
          sequence_no          (uint32)    Packet sequence number.
    '''
    controller: int
    event_index: int
    event_type: int
    event_access_granted: bool
    event_door: int
    event_direction: int
    event_card: int
    event_timestamp: datetime.datetime
    event_reason: int
    system_date: datetime.date
    system_time: datetime.time
    door_1_open: bool
    door_2_open: bool
    door_3_open: bool
    door_4_open: bool
    door_1_button: bool
    door_2_button: bool
    door_3_button: bool
    door_4_button: bool
    relays: int
    inputs: int
    system_error: int
    special_info: int
    sequence_no: int
